
#include"hash.h"

void *hsearch(void *base,int width,int size,unsigned key,match_fn match) {
          int pos,initial,found;
          void *element,*null;

          hashes = 0;
          null = NULL;
          initial = pos = hash(key,size);
          do {
             ++hashes;
             element = ((char *)base) + (pos * width);
             if(!(found = match(element))) {
                if(null != NULL) {
                   memswap(element,null,width);
                   return null;
                }
                return element;
             }
             if(found < 0 && null == NULL) null = element;
          }
          while(((1 << hashes) <= size) && ((pos = rehash(key,pos,size)) != initial));
          return NULL;
      }

void *kihsearch(void *base,int width,int size,unsigned key,match_fn match) {
          int pos,initial;
          void *element,*prev;

          hashes = 0;
          prev = NULL;
          initial = pos = hash(key,size);
          do {
             ++hashes;
             element = (char *)base + (pos * width);
             if(! match(element)) {
                if(prev != NULL) {
                   memswap(element,prev,width);
                   return prev;
                }
                return element;
             }
             prev = element;
          }
          while(((1 << hashes) <= size) && ((pos = kirehash(key,pos,size)) != initial));
          return NULL;
      }

void *ksihsearch(void *base,int width,int size,unsigned key,match_fn match) {
         void *element, *pelement;
         int prev,pprev,curr;

         pelement = NULL;
         hashes = 0;
         prev = -1;
         curr = hash(key,size);
         do {
            ++hashes;
            element = (char *)base + (curr * width);
            if(! match(element)) {
               if(pelement) {
                   memswap(pelement,element,width);
                   return pelement;
               }
               else return element;
            }
            pelement = element;
            pprev = prev;
            prev = curr;
            curr = ksirehash(key,curr,hashes,size);
        } while(((1 << hashes) <= size) && curr != prev && curr != pprev);
        return NULL;
   }

void **hresize(void **base,int width,int *size,int newsize,hsearch_fn search,key_fn key,match_fn match,fail_fn fail) {
          static void *newbase;
          void *element = *base;
          int oldsize = *size;

          newbase = hnew(newsize,width);
          if(newbase == NULL) return NULL;
          while(oldsize > 0) {
               void *newloc = search(newbase,width,newsize,key(element),match);
               if(newloc == NULL) {
                   if(fail(element)) {
                      errno = EHRESIZE_FAIL;
                      free(newbase);
                      return NULL;
                   }
               }
               else memcpy(newloc,element,width);
               oldsize--;
               element = (char *)element + width;
           }
           hdispose(*base);
           *base = newbase;
           *size = newsize;
           return base;
    }

#define CHAR_SHIFT_MAX 25
unsigned strsign(char *string) {
           unsigned sign = 0;

           while(*string) {
              sign ^= ((unsigned)(*string)) << (sign % CHAR_SHIFT_MAX);
              string++;
           }
           return sign;
        }

void memswap(void *loc0,void *loc1,int size) {
         while(size > sizeof(unsigned))  {
             unsigned temp = *(unsigned *)loc0;
             *(unsigned *)loc0 = *(unsigned *)loc1;
             *(unsigned *)loc1 = temp;
             (unsigned *)loc0++;
             (unsigned *)loc1++;
             size -= sizeof(unsigned);
         }
         while(size-- > 0) {
             char temp = *(char *)loc0;
             *(char *)loc0 = *(char *)loc1;
             *(char *)loc1 = temp;
             (char *)loc0++;
             (char *)loc1++;
         }
         return;
    }

void shdispose(syn_t *base,int size,freee_fn freee) {
        syn_t curr = *base, next = NULL;
        int index;

        for(index = 0;index < size;index++) {
            curr = base[index];
            while(curr) {
               next = curr->next;
               freee(curr->element);
               free(curr);
               curr = next;
            }
        }
        return;
     }

shptr_t shsearch(syn_t *base,int size,int type,unsigned key,match_fn match,shptr_t ptr) {
           int notfound = 1;

           ptr->index = (type) ? ohash(key,size) : hash(key,size);
           ptr->prev = NULL;
           ptr->curr = base[ptr->index];
           while(ptr->curr && (notfound = match(ptr->curr->element))) {
               if(type && (notfound > 0)) break;
               ptr->prev = ptr->curr;
               ptr->curr = ptr->curr->next;
           }
           if(! notfound) {
               if(! type) {
                   if(ptr->prev) {
                      ptr->prev->next = ptr->curr->next;
                      ptr->curr->next = base[ptr->index];
                      base[ptr->index] = ptr->curr;
                   }
              }
              return ptr;
           }
           return NULL;
        }

syn_t *shadd(syn_t *base,int size,unsigned key,void *element) {
            static syn_t new;

            new = calloc(1,sizeof(struct syn));
            if(new) {
                int pos = hash(key,size);
                new->element = element;
                new->next = base[pos];
                base[pos] = new;
                return base;
            }
            return NULL;
       }

syn_t *oshadd(syn_t *base,shptr_t ptr,void *element) {
            static syn_t new;

            new = calloc(1,sizeof(struct syn));
            if(new) {
                new->element = element;
                new->next = ptr->curr;
                if(ptr->prev) ptr->prev->next = new;
                else base[ptr->index] = new;
                ptr->curr = new;
                return base;
            }
            return NULL;
       }

syn_t *shrem(syn_t *base,shptr_t ptr) {
            syn_t temp;

            temp = base[ptr->index];
            if(temp) {
               ptr->prev = NULL;
               ptr->curr = base[ptr->index] = temp->next;
               free(temp->element);
               free(temp);
            }
            return base;
       }

syn_t *oshrem(syn_t *base,shptr_t ptr) {
            syn_t temp = ptr->curr;

            if(temp) {
               free(temp->element);
               ptr->curr = temp->next;
               if(ptr->prev) ptr->prev->next = temp->next;
               else base[ptr->index] = temp->next;
               free(temp);
               return base;
            }
            return NULL;
       }

shptr_t shseek(syn_t *base,int size,shptr_t ptr,int seek) {
            int index;

            switch(seek) {
                case SHSEEK_LAST : {
                                    index = size;
                                    while(((--index) >= 0) && ! base[index]);
                                    if(index < 0) return NULL;
                                    ptr->index = index;
                                    ptr->prev = NULL;
                                    ptr->curr = base[index];
                                    while(ptr->curr->next) {
                                        ptr->prev = ptr->curr;
                                        ptr->curr = ptr->curr->next;
                                    }
                                    return ptr;
                                    }
                case SHSEEK_NEXT :{
                                   ptr->prev = ptr->curr;
                                   ptr->curr = ptr->curr->next;
                                   if(ptr->curr) return ptr;
                                   index = ptr->index;
                                   break;
                                   }
                case SHSEEK_FIRST :{
                                    index = -1;
                                    break;
                                   }
                          default : return NULL;
           }
           while(1) {
                ++index;
                if(index >= size) return NULL;
                if(base[index] != NULL) {
                       ptr->index = index;
                       ptr->prev = NULL;
                       ptr->curr = base[index];
                       return ptr;
                }
          }
       }

syn_t **shresize(syn_t **base,int *size,int newsize,int type,key_fn key) {
            static syn_t *newbase;
            syn_t curr = NULL;
            int index = -1,pos;

            newbase = shnew(newsize);
            if(! newbase) return NULL;
            if(type) {
                int newindex = -1;
                syn_t prev = NULL;

                while(1) {
                    if(! curr) {
                        if((++index) >= *size) break;
                        curr = *base[index];
                        if(prev) prev->next = curr;
                    }
                    pos = ohash(key(curr->element),newsize);
                    if(pos != newindex) {
                         if(prev) prev->next = NULL;
                         prev = NULL;
                         newindex = pos;
                         newbase[pos] = curr;
                    }
                    prev = curr;
                    curr = curr->next;
                }
            }
            else {
                syn_t next = NULL;

                while(1) {
                   if(! curr) {
                      if((++index) > *size) break;
                      curr = *base[index];
                   }
                   next = curr->next;
                   pos = hash(key(curr->element),newsize);
                   curr->next = newbase[pos];
                   newbase[pos] = curr;
                   curr = next;
               }
            }
            free(*base);
            *base = newbase;
            *size = newsize;
            return base;
       }
